/* -----------------------------------------------------------------------------
 *    ___   __    __       _      __              __   __
 *   / _ | / /__ / /____ _| | /| / /__  ___  ____/ /  / /____ ____
 *  / __ |/  '_//  '_/ // / |/ |/ / _ `/ -_)/ __/ _ \/ __/ -_) __/
 * /_/ |_/_/\_\/_/\_\\_,_/|__/|__/\_,_/\__/ \__/_//_/\__/\__/_/
 *
 * 080824-1 (c)2009 Elektor International Media B.V.
 * -------------------------------------------------------------------------- */
/* -----------------------------------------------------------------------------
 * This file contains the source code of the function library to control
 * the Character LCD display.
 *
 * Notes
 *  The R/W pin is pulled low in hardware (connected to ground) so no reading
 * operation is permitted.
 * -------------------------------------------------------------------------- */

/* -----------------------------------------------------------------------------
 * Include
 * -------------------------------------------------------------------------- */
#include "LPC2103.h"
#include "HD44780.h"
#include "common.h"
#include "tools.h"


/* -----------------------------------------------------------------------------
 * LCD module pin connection description
 * -------------------------------------------------------------------------- */
enum LCD_GPIO_PIN_MAPPING {
	LCD_E  = 0x04000000, // GPIO P0.26
	LCD_RS = 0x00002000, // GPIO P0.13
	LCD_D4 = 0x00008000, // GPIO P0.15
	LCD_D5 = 0x00010000, // GPIO P0.16
	LCD_D6 = 0x00020000, // GPIO P0.17
	LCD_D7 = 0x00040000, // GPIO P0.18

	LCD_DUMMY_PIN
};

/* -----------------------------------------------------------------------------
 * LCD controller controller registers, commands and addresses
 * -------------------------------------------------------------------------- */
// Function Set
#define LCD_REG_FUNCTION_SET  0x20
#define LCD_CMD_DATA_8BITS    0x10
#define LCD_CMD_DATA_4BITS    0x00
#define LCD_CMD_ONE_LINE      0x00
#define LCD_CMD_TWO_LINES     0x08
#define LCD_CMD_5x7_FONT      0x00
#define LCD_CMD_5x10_FONT     0x04
// Display ON/OFF control
#define LCD_REG_ONOFF_CTRL    0x08
#define LCD_CMD_DISPLAY_ON    0x04
#define LCD_CMD_DISPLAY_OFF   0x00
#define LCD_CMD_CURSOR_ON     0x02
#define LCD_CMD_CURSOR_OFF    0x00
#define LCD_CMD_CURSOR_BLINK  0x01
// Cursor and display shift
#define LCD_REG_SHIFT_CTRL    0x10
#define LCD_CMD_MOVE_CURSOR   0x00
#define LCD_CMD_MOVE_DISPLAY  0x08
#define LCD_CMD_MOVE_RIGHT    0x01
#define LCD_CMD_MOVE_LEFT     0x00
// Entry mode set
// Cursor and display shift
#define LCD_REG_ENTRY_MODE    0x04
#define LCD_CMD_ADDRESS_INC   0x02
#define LCD_CMD_ADDRESS_DEC   0x00
#define LCD_CMD_SHIFT_DISPLAY 0x01
// direct rendering functions
#define LCD_CMD_CLEAR_DISPLAY 0x01
#define LCD_CMD_GO_HOME       0x02
// CGRAM (Character Generator RAM) base address
#define LCD_REG_CGRAM         0x40
// DDRAM (Display Data RAM) base address
#define LCD_REG_DGRAM         0x80

/* -----------------------------------------------------------------------------
 * LCD module delay and timing features
 * -------------------------------------------------------------------------- */
// according to datasheet DEM16217SYPH-PY p.13 we have the folowing
// wait for more than 30ms minimum after VDD rises to 4.5V
#define DELAY_DCDC_STABILIZE 50000
// minimum E cycle is 500ns
#define DELAY_DATA_LATCH     1
// wait for more than 39us after a command is set
#define DELAY_FUNCTION_SET   50
// wait for more than 1.53ms after display is cleared
#define DELAY_CLEAR_DGRAM    1600




/* -----------------------------------------------------------------------------
 * Private Functions
 * -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/**@author Antoine Authier
 * @date   20090306
 * @brief  write a nibble (4 bits) to the LCD MSB data port
 * @param[in] byte contains in its LSB the nibble to send to the LCD
 * @note   byte MSB are not used and can contain anything
 */
static void
LCD_send_nibble (char byte)
{
	// clear the data nibble already at the GPIO
	IOCLR = LCD_D4 | LCD_D5 | LCD_D6 | LCD_D7;

	// write high nibble to data pins
	if (byte & MASK_BIT3)
		IOSET = LCD_D7;
	if (byte & MASK_BIT2)
		IOSET = LCD_D6;
	if (byte & MASK_BIT1)
		IOSET = LCD_D5;
	if (byte & MASK_BIT0)
		IOSET = LCD_D4;
}

/* -------------------------------------------------------------------------- */
/**@author Antoine Authier
 * @date   20090306
 * @brief  strobe LCD Enable signal to latch data to the LCD controller
 * @param[in] delay in micro seconds to wait for a correct data latching
 * @note   refer to LCD controller module datasheet for typical delays!
 */
static void
LCD_data_strobe (void)
{
	// Enable signal goes logic High
	IOSET = LCD_E;
	// wait for data latching
	delay_us(DELAY_DATA_LATCH);
	// Enable signal goes logic Low
	IOCLR = LCD_E;
}

/* -------------------------------------------------------------------------- */
/**@author Steffen Graf / Antoine Authier
 * @date   20090306
 * @brief  send a data byte to the LCD
 * @param[in] byte contains the data to send to the LCD
 */
static void
LCD_send_data (char byte)
{
	// set RS to select data register
	IOSET = LCD_RS;

	// write high nibble to data pins
	LCD_send_nibble((byte >> 4) & 0x0F);
	// latch data in the LCD controller
	LCD_data_strobe();

	// write low nibble to data pins
	LCD_send_nibble(byte & 0x0F);
	// latch data in the LCD controller
	LCD_data_strobe();
}

/* -------------------------------------------------------------------------- */
/**@author Steffen Graf / Antoine Authier
 * @date   20090306
 * @brief  send a command to the LCD
 * @param[in] byte contains the command to send to the LCD
 */
static void
LCD_send_cmd (char byte)
{
	// set RS to select data register
	IOCLR = LCD_RS;

	// write high nibble to data pins
	LCD_send_nibble((byte >> 4) & 0x0F);
	// latch data in the LCD controller
	LCD_data_strobe();

	// write low nibble to data pins
	LCD_send_nibble(byte & 0x0F);
	// latch data in the LCD controller
	LCD_data_strobe();

	delay_us(DELAY_FUNCTION_SET);
}



/* -----------------------------------------------------------------------------
 * Public Functions
 * -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
/**@author Steffen Graf / Antoine Authier
 * @date   20090306
 * @brief  setup communication and initialize LCD hardware module
 */
void
LCD_initialize (void)
{
	unsigned char byte;

	// setup MCU GPIOs, D[4..7], E and RS are outputs
	IODIR |= LCD_D4 | LCD_D5 | LCD_D6 | LCD_D7 | LCD_E | LCD_RS;

	// wait for LCD supply to stabilize
	delay_us(DELAY_DCDC_STABILIZE);

	// configure LCD interface data lenght to 4 bits mode
	IOCLR = LCD_RS;
	byte = LCD_REG_FUNCTION_SET | LCD_CMD_DATA_4BITS;
	LCD_send_nibble((byte >> 4) & 0x0F);
	LCD_data_strobe();

	// set LCD configuration: 2 lines, font 5x7 dots, still 4 bits mode
	// CAUTION always match the number of lines with
	// the global define LCD_NUMBER_OF_LINES
	byte = LCD_REG_FUNCTION_SET | LCD_CMD_DATA_4BITS;
	byte |= LCD_CMD_TWO_LINES | LCD_CMD_5x7_FONT;
	LCD_send_cmd(byte);

	// set display ON/OFF control: display ON, no cursor
	byte = LCD_REG_ONOFF_CTRL | LCD_CMD_DISPLAY_OFF | LCD_CMD_CURSOR_OFF;
	LCD_send_cmd(byte);

	// clear display and go home
	LCD_clear();

	// set cursor and display shift: 
	byte = LCD_REG_ENTRY_MODE | LCD_CMD_ADDRESS_INC;
	LCD_send_cmd(byte);

	// display on
	byte = LCD_REG_ONOFF_CTRL | LCD_CMD_DISPLAY_ON | LCD_CMD_CURSOR_OFF;
	LCD_send_cmd(byte);
}

/* -------------------------------------------------------------------------- */
/**@author Steffen Graf
 * @date   20090306
 * @brief  clear the display and place cursor at home
 */
void
LCD_clear (void)
{
	// clear display
	LCD_send_cmd(LCD_CMD_CLEAR_DISPLAY);
	delay_us(DELAY_CLEAR_DGRAM);

	// place cursor at home
	LCD_send_cmd(LCD_CMD_GO_HOME);
}

/* -------------------------------------------------------------------------- */
/**@author Steffen Graf / Antoine Authier
 * @date   20090306
 * @brief  move cursor on the LCD to (x, y) location
 * @param[in] y is the vertical position (ID of the line)
 * @param[in] x is the horizontal position (ID of the block in the line)
 * @note   Both IDs ranges starts from 0 to LCD_NUMBER_OF_LINES and LCD_LINE_WIDTH
 *         minus one. Please use defines for the line ID.
 */
void
LCD_move_cursor (unsigned char y, unsigned char x)
{
	// CAUTION the line addresses depends on your module specification
	//         please check your datasheet!
	static unsigned char line_addreses[4] = {0x00, 0x40, 0x14, 0x54};
	unsigned char position;

	// calculate new DDRAM pointer address according to the location
	position = LCD_REG_DGRAM + line_addreses[y] + x;
	// set the pointer
	LCD_send_cmd(position);
}

/* -------------------------------------------------------------------------- */
/**@author Steffen Graf / Antoine Authier
 * @date   20090306
 * @brief  display a text on the display
 * @param[in] string contains the text to display and must be NULL terminated
 */
void
LCD_print_text (char *string)
{
	// send each caracter of the string until the NULL character
	while (*string != 0x00) {
		LCD_send_data(*string);
		string++;
	}
}

/* -------------------------------------------------------------------------- */
/**@author Antoine Authier
 * @date   20090306
 * @brief  display a character on the display
 * @param[in] character to display
 * @note   this is a wrapper for the high level LCD functions
 */
void
LCD_print_character (char character)
{
	LCD_send_data(character);
}

/* -------------------------------------------------------------------------- */
/**@author Steffen Graf
 * @date   20090306
 * @brief  display an integer value with 1 figures on the display
 * @param[in] value contains the (short :) integer to display
 */
void
LCD_int1 (unsigned short value)
{
	LCD_send_data('0' + value);
}

/* -------------------------------------------------------------------------- */
/**@author Steffen Graf
 * @date   20090306
 * @brief  display an integer value with 2 figures on the display
 * @param[in] value contains the (short :) integer to display
 */
void
LCD_int2 (unsigned short value)
{
	unsigned char byte;

	byte = value / 10;
	value = value - byte * 10;
	LCD_send_data('0' + byte);

	LCD_send_data('0' + value);
}

/* -------------------------------------------------------------------------- */
/**@author Steffen Graf
 * @date   20090306
 * @brief  display an integer value with 3 figures on the display
 * @param[in] value contains the (short :) integer to display
 */
void
LCD_int3 (unsigned short value)
{
	unsigned char byte;

	byte = value / 100;
	value = value - byte * 100;
	LCD_send_data('0' + byte);

	byte = value / 10;
	value = value - byte * 10;
	LCD_send_data('0' + byte);

	LCD_send_data('0' + value);
}

/* -------------------------------------------------------------------------- */
/**@author Steffen Graf
 * @date   20090306
 * @brief  display an integer value with 4 figures on the display
 * @param[in] value contains the (short :) integer to display
 */
void
LCD_int4 (unsigned short value)
{
	unsigned char byte;

	byte = value / 1000;
	value = value - byte * 1000;
	LCD_send_data('0' + byte);

	byte = value / 100;
	value = value - byte * 100;
	LCD_send_data('0' + byte);

	byte = value / 10;
	value = value - byte * 10;
	LCD_send_data('0' + byte);

	LCD_send_data('0' + value);
}

/* -------------------------------------------------------------------------- */
/**@author Steffen Graf
 * @date   20090306
 * @brief  display an integer value with 5 figures on the display
 * @param[in] value contains the (short :) integer to display
 */
void
LCD_int5 (unsigned short value)
{
	unsigned char byte;

	byte = value / 10000;
	value = value - byte * 10000;
	LCD_send_data('0' + byte);

	byte = value / 1000;
	value = value - byte * 1000;
	LCD_send_data('0' + byte);

	byte = value / 100;
	value = value - byte * 100;
	LCD_send_data('0' + byte);

	byte = value / 10;
	value = value - byte * 10;
	LCD_send_data('0' + byte);

	LCD_send_data('0' + value);
}
